import datetime, pprint

log_printer = pprint.PrettyPrinter(compact=False, width=40).pprint


def log(
    message: str,
    level: str = "",
    color: bool = True,
    datetime_format: str = "%Y-%m-%d %H:%M:%S",
):
    timestamp = datetime.datetime.now().strftime(datetime_format)

    level = level.lower()

    color_code = ""
    if color:
        if level == "" or level == "message":
            pass
        elif level == "ok" or level == "success":
            color_code = "\u001b[32;1m"
        elif level == "info":
            color_code = "\u001b[34;1m"
        elif level == "warn":
            color_code = "\u001b[33;1m"
        elif level == "error":
            color_code = "\u001b[31;1m"

    level = "(" + level.upper() + ") " if level != "" else ""

    if isinstance(message, str):
        print(
            "{color_code}[{timestamp}] {level}{message}{end_color}".format(
                color_code=color_code,
                timestamp=timestamp,
                level=level,
                message=message,
                end_color="\033[0m" if color else "",
            )
        )

    else:

        return_val = "{color_code}[{timestamp}] ---[ Printing {var_type} ]---\n".format(
            color_code=color_code, timestamp=timestamp, var_type=type(message)
        )
        return_val += pprint.pformat(message) + "\n"
        return_val += "[{timestamp}] ---[ End      {var_type} ]---{end_color}".format(
            timestamp=timestamp,
            var_type=type(message),
            end_color="\033[0m" if color else "",
        )
        print(return_val)
        return

    return


import functools, time, math, warnings

# Credit to https://realpython.com/primer-on-python-decorators/#timing-functions
def timer(func):
    """Print the runtime of the decorated function"""

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()  # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()  # 2
        run_time = end_time - start_time  # 3
        print(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        return value

    return wrapper_timer


# Credit to https://realpython.com/primer-on-python-decorators/#debugging-code
def debug(func):
    """Print the function signature and return value"""

    @functools.wraps(func)
    def wrapper_debug(*args, **kwargs):
        args_repr = [repr(a) for a in args]  # 1
        kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]  # 2
        signature = ", ".join(args_repr + kwargs_repr)  # 3
        print(f"Calling {func.__name__}({signature})")
        value = func(*args, **kwargs)
        print(f"{func.__name__!r} returned {value!r}")  # 4
        return value

    return wrapper_debug


def count_calls(func):
    @functools.wraps(func)
    def wrapper_count_calls(*args, **kwargs):
        wrapper_count_calls.num_calls += 1
        print(f"Call {wrapper_count_calls.num_calls} of {func.__name__!r}")
        return func(*args, **kwargs)

    wrapper_count_calls.num_calls = 0
    return wrapper_count_calls


def singleton(cls):
    """Make a class a Singleton class (only one instance)"""

    @functools.wraps(cls)
    def wrapper_singleton(*args, **kwargs):
        if not wrapper_singleton.instance:
            wrapper_singleton.instance = cls(*args, **kwargs)
        return wrapper_singleton.instance

    wrapper_singleton.instance = None
    return wrapper_singleton


# Credit to https://wiki.python.org/moin/PythonDecoratorLibrary#Retry
def retry(tries, delay=3, backoff=2):
    """Retries a function or method until it returns True.

    delay sets the initial delay in seconds, and backoff sets the factor by which
    the delay should lengthen after each failure. backoff must be greater than 1,
    or else it isn't really a backoff. tries must be at least 0, and delay
    greater than 0."""

    if backoff <= 1:
        raise ValueError("backoff must be greater than 1")

    tries = math.floor(tries)
    if tries < 0:
        raise ValueError("tries must be 0 or greater")

    if delay <= 0:
        raise ValueError("delay must be greater than 0")

    def deco_retry(f):
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay  # make mutable

            rv = f(*args, **kwargs)  # first attempt
            while mtries > 0:
                if rv is True:  # Done on success
                    return True

                mtries -= 1  # consume an attempt
                time.sleep(mdelay)  # wait...
                mdelay *= backoff  # make future wait longer

                rv = f(*args, **kwargs)  # Try again

            return False  # Ran out of tries :-(

        return f_retry  # true decorator -> decorated function

    return deco_retry  # @retry(arg[, ...]) -> true decorator


# Credit to https://wiki.python.org/moin/PythonDecoratorLibrary#Smart_deprecation_warnings_.28with_valid_filenames.2C_line_numbers.2C_etc..29
def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter("always", DeprecationWarning)
        warnings.warn_explicit(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
            filename=func.__code__.co_filename,
            lineno=func.__code__.co_firstlineno + 1,
        )
        return func(*args, **kwargs)

    return new_func



from decorators import *
import time


@timer
def take_some_time():
    time.sleep(1)
    print("hello")


@debug
def debug_func(message):
    print(message)
    return True


@count_calls
def call_twice():
    print("Hi")


@deprecated
def deprecated_func():
    print("This is deprecated")


print()

take_some_time()

print()

debug_func("hello, world")

call_twice()
call_twice()

print()

warnings.simplefilter("always", DeprecationWarning)
deprecated_func()

print()


from logging import log

log("test - no level")

log("test ok", "ok")
log("test success", "success")
log("test info", "info")
log("test warn", "warn")
log("test error", "error")
log("test message", "message")

log(
    {
        "test": "message",
        "a": "b",
        "c": "d",
        "test_array": ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k"],
    },
    "warn",
)
log([1, 2, 3, 4, 5], "info")
